﻿/*
  SpmImport Plugin
  Copyright (C) 2015 Dominik Reichl <dominik.reichl@t-online.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.Drawing;
using System.IO;
using System.IO.Packaging;
using System.Xml;
using System.Resources;
using System.Diagnostics;

using KeePass;
using KeePass.DataExchange;
using KeePass.Resources;

using KeePassLib;
using KeePassLib.Interfaces;
using KeePassLib.Utility;

namespace SpmImport
{
	public sealed class SpmXps : FileFormatProvider
	{
		public override bool SupportsImport { get { return true; } }
		public override bool SupportsExport { get { return false; } }

		public override string FormatName
		{
			get { return "Steganos Password Manager XPS/OXPS"; }
		}

		public override string DefaultExtension { get { return "xps|oxps"; } }
		public override string ApplicationGroup { get { return KPRes.Plugins; } }

		public override Image SmallIcon
		{
			get
			{
				Image img = null;

				ResourceManager rm = Program.Resources;
				if(rm != null)
					img = (rm.GetObject("B16x16_Imp_Steganos") as Image);

				if(img != null) return img;
				else { Debug.Assert(false); }

				return base.SmallIcon;
			}
		}

		private static readonly string[] g_vGroupPrefixes = new string[] {
			"Category:", "Kategorie:"
		};

		private static readonly string[] g_vSubGroupPrefixes = new string[] {
			"Sub-category:", "Unterkategorie:"
		};

		private static readonly string[] g_vEntryPrefixes = new string[] {
			"Password Entry:", "Passwort-Eintrag:"
		};

		private static Dictionary<string, string> g_dFields = null;

		public override void Import(PwDatabase pwStorage, Stream sInput,
			IStatusLogger slLogger)
		{
			if(g_dFields == null)
			{
				Dictionary<string, string> d = new Dictionary<string, string>();

				d["User name:"] = PwDefs.UserNameField;
				d["Benutzername:"] = PwDefs.UserNameField;

				d["Password:"] = PwDefs.PasswordField;
				d["Passwort:"] = PwDefs.PasswordField;

				d["URL:"] = PwDefs.UrlField;

				d["About:"] = PwDefs.NotesField;
				d["Info:"] = PwDefs.NotesField;

				g_dFields = d;
			}

			Package pk = Package.Open(sInput, FileMode.Open, FileAccess.Read);
			if(pk == null)
				throw new Exception("Failed to open XPS/OXPS file!");

			try { ImportPackage(pk, pwStorage); }
			finally { pk.Close(); }
		}

		private void ImportPackage(Package pk, PwDatabase pd)
		{
			PackagePart ppDocIndex = pk.GetPart(GetPackageUri(pk,
				"/FixedDocumentSequence.fdseq", string.Empty));
			if(ppDocIndex == null)
				throw new Exception("Fixed document sequence not found!");

			Stream sDocIndex = GetPartStream(ppDocIndex);
			try
			{
				XmlDocument xd = new XmlDocument();
				xd.Load(sDocIndex);

				foreach(XmlNode xn in xd.DocumentElement)
				{
					if(xn.Name == "DocumentReference")
					{
						try
						{
							string strDoc = xn.Attributes["Source"].Value;
							if(!string.IsNullOrEmpty(strDoc))
							{
								PackagePart ppDoc = pk.GetPart(
									GetPackageUri(pk, strDoc, string.Empty));

								ImportDocument(pk, ppDoc, pd);
							}
							else { Debug.Assert(false); }
						}
						catch(Exception) { Debug.Assert(false); }
					}
				}
			}
			finally { sDocIndex.Close(); }
		}

		private void ImportDocument(Package pk, PackagePart ppDoc, PwDatabase pd)
		{
			PwGroup[] vGroups = new PwGroup[2] { pd.RootGroup, null };
			PwEntry pe = null;
			string strField = PwDefs.NotesField;

			Stream sDoc = GetPartStream(ppDoc);
			try
			{
				XmlDocument xd = new XmlDocument();
				xd.Load(sDoc);

				foreach(XmlNode xn in xd.DocumentElement)
				{
					if(xn.Name == "PageContent")
					{
						try
						{
							string strPage = xn.Attributes["Source"].Value;
							if(!string.IsNullOrEmpty(strPage))
							{
								PackagePart ppPage = pk.GetPart(
									GetPackageUri(pk, strPage, ppDoc.Uri.ToString()));

								ImportPage(pk, ppPage, pd, vGroups, ref pe,
									ref strField);
							}
							else { Debug.Assert(false); }
						}
						catch(Exception) { Debug.Assert(false); }
					}
				}
			}
			finally { sDoc.Close(); }
		}

		private void ImportPage(Package pk, PackagePart ppPage, PwDatabase pd,
			PwGroup[] vGroups, ref PwEntry pe, ref string strField)
		{
			Stream sPage = GetPartStream(ppPage);
			try
			{
				XmlDocument xd = new XmlDocument();
				xd.Load(sPage);

				XmlElement xe = xd.DocumentElement;
				XmlNodeList lGlyphs = xe.GetElementsByTagName("Glyphs");

				foreach(XmlNode xn in lGlyphs)
				{
					try
					{
						string str = xn.Attributes["UnicodeString"].Value;

						ImportText(str, pd, vGroups, ref pe, ref strField);
					}
					catch(Exception) { Debug.Assert(false); }
				}
			}
			finally { sPage.Close(); }
		}

		private void ImportText(string strText, PwDatabase pd, PwGroup[] vGroups,
			ref PwEntry pe, ref string strField)
		{
			if(strText == null) { Debug.Assert(false); return; }

			string str = strText.Trim();

			// Ignore page numbers
			if(Regex.IsMatch(str, "^\\d+\\s+/\\s+\\d+$")) return;

			foreach(string strGroupPrefix in g_vGroupPrefixes)
			{
				if(str.StartsWith(strGroupPrefix, StrUtil.CaseIgnoreCmp))
				{
					string strGroup = str.Substring(strGroupPrefix.Length).Trim();
					if(strGroup.Length > 0)
						vGroups[0] = pd.RootGroup.FindCreateGroup(strGroup, true);
					else { Debug.Assert(false); vGroups[0] = pd.RootGroup; }

					vGroups[1] = null;
					return;
				}
			}

			foreach(string strSubGroupPrefix in g_vSubGroupPrefixes)
			{
				if(str.StartsWith(strSubGroupPrefix, StrUtil.CaseIgnoreCmp))
				{
					string strGroup = str.Substring(strSubGroupPrefix.Length).Trim();
					if(strGroup.Length > 0)
						vGroups[1] = vGroups[0].FindCreateGroup(strGroup, true);
					else { Debug.Assert(false); vGroups[1] = null; }

					return;
				}
			}

			foreach(string strEntryPrefix in g_vEntryPrefixes)
			{
				if(str.StartsWith(strEntryPrefix, StrUtil.CaseIgnoreCmp))
				{
					PwGroup pg = (vGroups[1] ?? vGroups[0]);

					pe = new PwEntry(true, true);
					pg.AddEntry(pe, true);

					str = str.Substring(strEntryPrefix.Length);
					strField = PwDefs.TitleField;
				}
			}

			foreach(KeyValuePair<string, string> kvpField in g_dFields)
			{
				if(str.StartsWith(kvpField.Key, StrUtil.CaseIgnoreCmp))
				{
					str = str.Substring(kvpField.Key.Length);
					strField = kvpField.Value;
				}
			}

			str = str.Replace(@"<-N/L-/>", MessageService.NewLine);
			str = str.Trim();

			// The following strings (that are ignored) must be checked
			// *after* the group parsing, because for instance the English
			// "keychain" by default appears in a category name
			if(str.StartsWith("Schl\u00FCsselbund", StrUtil.CaseIgnoreCmp)) return;
			if(str.IndexOf(@"'s keychain", StrUtil.CaseIgnoreCmp) >= 0) return;

			if(pe == null) return;

			bool bNotes = (strField == PwDefs.NotesField);
			ImportUtil.AppendToField(pe, strField, str, pd,
				(bNotes ? MessageService.NewLine : " "), !bNotes);
		}

		private static Stream GetPartStream(PackagePart pp)
		{
			using(Stream s = pp.GetStream(FileMode.Open, FileAccess.Read))
			{
				MemoryStream ms = new MemoryStream();
				MemUtil.CopyStream(s, ms);
				byte[] pb = ms.ToArray();
				ms.Close();

				return new MemoryStream(pb, false);
			}
		}

		private static Uri GetPackageUri(Package pk, string strRelUri,
			string strRelBaseFile)
		{
			string strUri = strRelUri;

			// XPS stores absolute paths, OXPS stores relative paths
			if(!strRelUri.StartsWith("/"))
				strUri = UrlUtil.GetFileDirectory(strRelBaseFile, false,
					false) + "/" + strRelUri;

			return new Uri(strUri, UriKind.Relative);
		}
	}
}
